LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;
USE ieee.numeric_std.ALL;

library work;
USE work.constants.ALL;

-- Arithmetic Logic Unit
-- Does not know how to deal with immediates, only raw values
ENTITY alu IS
	PORT(
		-- Inputs
		-------------
		clk		: IN  STD_LOGIC;
		op_enable	: IN  STD_LOGIC;
		opcode		: IN  STD_LOGIC_VECTOR(OPCODE_LEN DOWNTO 0);
		reg1_val	: IN  STD_LOGIC_VECTOR(XLENM1 DOWNTO 0); 
		reg2_val	: IN  STD_LOGIC_VECTOR(XLENM1 DOWNTO 0); 

		-- Outputs
		-------------
		dest_val	: OUT STD_LOGIC_VECTOR(XLENM1 DOWNTO 0)
	);
END alu;

ARCHITECTURE rtl OF alu IS

-- Internal signals, representing each calculation
SIGNAL add_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL sub_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL sll_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL slt_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL sltu_out: STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL xor_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL srl_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL sra_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL or_out  : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL and_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL raw_mul_uu : STD_LOGIC_VECTOR(XLENM1+XLEN DOWNTO 0);
SIGNAL raw_mul_su : STD_LOGIC_VECTOR(XLENM1+XLEN DOWNTO 0);
SIGNAL raw_mul_ss : STD_LOGIC_VECTOR(XLENM1+XLEN DOWNTO 0);
SIGNAL mul_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL mulh_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL mulhsu_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL mulhu_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL div_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL divu_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL rem_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL remu_out : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
SIGNAL shift_amt : STD_LOGIC_VECTOR(5 DOWNTO 0);

-- Useful constants; 1, Unsigned and Signed max and min
CONSTANT VAL_ONE : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0) := x"00000001";
CONSTANT VAL_UMAX : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0) := (OTHERS => '1');
CONSTANT VAL_UMIN : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0) := (OTHERS => '0');
CONSTANT VAL_SMAX : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0) := (XLENM1 => '0',OTHERS => '1');
CONSTANT VAL_SMIN : STD_LOGIC_VECTOR(XLENM1 DOWNTO 0) := (XLENM1 => '1',OTHERS => '0');


BEGIN

--Mask away extra shift values, since our inputs are register values rather than register addresses
shift_amt <= reg2_val(5 DOWNTO 0);

--RV32I base; +, -, ||, &&, XOR, Logical Shift Left (SLL), Logical shift right (SRL), 
-- arithmetic shift right (SRA), Set Less Than unsigned (SLTU), Set Less Than (SLT)
-----------------------------
add_out <= reg1_val +   reg2_val;
sub_out <= reg1_val -   reg2_val;
or_out  <= reg1_val OR  reg2_val;
and_out <= reg1_val AND reg2_val;
xor_out <= reg1_val XOR reg2_val;
sll_out <= std_logic_vector(shift_left(unsigned(reg1_val), to_integer(unsigned(shift_amt))));
srl_out <= std_logic_vector(shift_right(unsigned(reg1_val), to_integer(unsigned(shift_amt))));
sra_out <= std_logic_vector(shift_right(signed(reg1_val), to_integer(unsigned(shift_amt))));
sltu_out <= VAL_ONE when unsigned(reg1_val) < unsigned(reg2_val) else VAL_ZERO;
slt_out <= VAL_ONE when signed(reg1_val) < signed(reg2_val) else VAL_ZERO;


--M extension
-----------------------------
-- Pre-calculate for all sign variants
raw_mul_uu <= std_logic_vector(unsigned(reg1_val) * unsigned(reg2_val));
raw_mul_su <= std_logic_vector(signed(reg1_val) * unsigned(reg2_val));
raw_mul_ss <= std_logic_vector(signed(reg1_val) * signed(reg2_val));
--MUL (multiply)
mul_out <= raw_mul_ss(XLENM1 DOWNTO 0);
--MULH (Multiply High)
mulh_out <= raw_mul_ss(XLENM1+XLEN DOWNTO XLEN);
--MULHSU (Multiply High; Signed by unsigned)
mulh_out <= raw_mul_su(XLENM1+XLEN DOWNTO XLEN);
--MULHU (Multiply High; unsigned)
mulh_out <= raw_mul_uu(XLENM1+XLEN DOWNTO XLEN);
--DIV (Divide)
div : PROCESS(clk) BEGIN
IF(VAL_UMIN = reg2_val) THEN
	div_out <= VAL_UMAX;
ELSIF (VAL_SMAX = reg1_val AND VAL_UMAX = reg2_val) THEN
	div_out <= VAL_SMIN;
ELSE
	div_out <= std_logic_vector(signed(reg1_val) / signed(reg2_val));
END IF;
END PROCESS;
--DIVU (Divide unsigned)
divu : PROCESS(clk) BEGIN
IF(VAL_UMIN = reg2_val) THEN
	divu_out <= VAL_UMAX;
ELSE
	div_out <= std_logic_vector(unsigned(reg1_val) / unsigned(reg2_val));
END IF;
END PROCESS;
--REM (Remainder)
rem_inst : PROCESS(clk) BEGIN
IF(VAL_UMIN = reg2_val) THEN
	rem_out <= reg1_val;
ELSIF (VAL_SMAX = reg1_val AND VAL_UMAX = reg2_val) THEN
	rem_out <= reg1_val;
ELSE
	rem_out <= std_logic_vector(signed(reg1_val) rem signed(reg2_val));
END PROCESS;
--REMU (Remainder unsigned)
remu_inst : PROCESS(clk) BEGIN
IF(VAL_UMIN = reg2_val) THEN
	remu_out <= VAL_UMIN;
ELSE
	remu_out <= std_logic_vector(signed(reg1_val) rem signed(reg2_val));
END PROCESS;


-- Output Implementation
-----------------------------
mux : PROCESS(clk) BEGIN
IF (rising_edge(clk)) THEN
	-- If we aren't supposed to be outputting a value, then don't
	IF '0' = op_enable THEN
		dest_val <= (OTHERS => '0');
	ELSE
		-- Check against internal opcode which calculation to output
		-- Not ideal, since multiplication/division is expensive compared
		-- to everything else, but this is a Proof of Concept
		CASE opcode IS
			WHEN OP_ADD    => dest_val <= add_out;
			WHEN OP_SUB    => dest_val <= sub_out;
			WHEN OP_SLL    => dest_val <= sll_out;
			WHEN OP_SLT    => dest_val <= slt_out;
			WHEN OP_SLTU   => dest_val <= sltu_out;
			WHEN OP_XOR    => dest_val <= xor_out;
			WHEN OP_SRL    => dest_val <= srl_out;
			WHEN OP_SRA    => dest_val <= sra_out;
			WHEN OP_OR     => dest_val <= or_out;
			WHEN OP_AND    => dest_val <= and_out;
			WHEN OP_MUL    => dest_val <= mul_out;
			WHEN OP_MULH   => dest_val <= mulh_out;
			WHEN OP_MULHSU => dest_val <= mulhsu_out;
			WHEN OP_MULHU  => dest_val <= mulhu_out;
			WHEN OP_DIV    => dest_val <= div_out;
			WHEN OP_DIVU   => dest_val <= divu_out;
			WHEN OP_REM    => dest_val <= rem_out;
			WHEN OP_REMU   => dest_val <= remu_out;
			WHEN OTHERS => dest_val <= (OTHERS => '0');
		END CASE;
	END IF;
END IF;
END PROCESS;

END rtl;
