LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;
USE ieee.numeric_std.ALL;

library riscv32;
USE riscv32.constants.ALL;

-- RV32G Instruction decoder
ENTITY inst_decoder IS
	PORT(
		--Inputs
		--**********************************************************
		instruction : IN  STD_LOGIC_VECTOR(REGLENM1 DOWNTO 0);
		clk : IN STD_LOGIC;

		--Outputs
		--**********************************************************
		-- General use
		--These are wired to all outputs
		destination	: OUT STD_LOGIC_VECTOR(REGLENM1 DOWNTO 0);
		reg1		: OUT STD_LOGIC_VECTOR(REGLENM1 DOWNTO 0);
		reg2		: OUT STD_LOGIC_VECTOR(REGLENM1	DOWNTO 0);
		-- Use Immediate instead of Reg2
		immediate_flag  : OUT STD_LOGIC;
		immediate	: OUT STD_LOGIC_VECTOR(XLENM1   DOWNTO 0);
		--Integer load
		xload_flag	: OUT STD_LOGIC;
		xload_addr	: OUT STD_LOGIC_VECTOR(REGLENM1 DOWNTO 0);
		xload_mask	: OUT STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
		xload_offset	: OUT STD_LOGIC_VECTOR(IMM_I_LEN DOWNTO 0);
		--Floating point load
		fload_flag	: OUT STD_LOGIC;
		fload_addr	: OUT STD_LOGIC_VECTOR(REGLENM1 DOWNTO 0);
		fload_mask	: OUT STD_LOGIC_VECTOR(FLENM1 DOWNTO 0);
		fload_offset	: OUT STD_LOGIC_VECTOR(IMM_I_LEN DOWNTO 0);
		--AUIPC
		auipc_flag	: OUT STD_LOGIC;
		auipc_imm	: OUT STD_LOGIC_VECTOR(IMM_U_LEN DOWNTO 0);
		--LUI
		lui_flag	: OUT STD_LOGIC;
		lui_imm		: OUT STD_LOGIC_VECTOR(IMM_U_LEN DOWNTO 0);
		--Integer store
		xstore_flag	: OUT STD_LOGIC;
		xstore_imm	: OUT STD_LOGIC_VECTOR(IMM_I_LEN DOWNTO 0);
		xstore_mask	: OUT STD_LOGIC_VECTOR(XLENM1 DOWNTO 0);
		--Floating point store
		fstore_flag	: OUT STD_LOGIC;
		fstore_imm	: OUT STD_LOGIC_VECTOR(IMM_I_LEN DOWNTO 0);
		fstore_mask	: OUT STD_LOGIC_VECTOR(FLENM1 DOWNTO 0);
		--Atomics
		amo_flag	: OUT STD_LOGIC;
		amo_funct7	: OUT STD_LOGIC_VECTOR(FUNCT7_LEN DOWNTO 0);
		--General integer operations
		op_flag		: OUT STD_LOGIC;
		op_op		: OUT STD_LOGIC_VECTOR(OPCODE_LEN DOWNTO 0);
		--Floating Point (single and half precision) operations
		flop_flag	: OUT STD_LOGIC;
		hflop_flag	: OUT STD_LOGIC;
		flop_r3		: OUT STD_LOGIC_VECTOR(REGLENM1 DOWNTO 0);
		flop_rm		: OUT STD_LOGIC_VECTOR(FUNCT3_LEN DOWNTO 0); 
		flop_op		: OUT STD_LOGIC_VECTOR(FLOAT_OP_LENM1 DOWNTO 0); 
		--Jump operations
		jalr_flag	: OUT STD_LOGIC;
		jalr_base	: OUT STD_LOGIC_VECTOR(REGLENM1 DOWNTO 0);
		jalr_offset	: OUT STD_LOGIC_VECTOR(IMM_I_LEN DOWNTO 0); 
		jal_flag	: OUT STD_LOGIC;
		jal_offset	: OUT STD_LOGIC_VECTOR(IMM_J_LEN DOWNTO 0);
		--Branch operations
		branch_flag	: OUT STD_LOGIC;
		branch_offset   : OUT STD_LOGIC_VECTOR(IMM_B_LEN DOWNTO 0);
		branch_op	: OUT STD_LOGIC_VECTOR(FUNCT3_LEN DOWNTO 0);
		--System 
		ecall_flag	: OUT STD_LOGIC;
		ebreak_flag	: OUT STD_LOGIC;
		nto_flag	: OUT STD_LOGIC;
		sto_flag	: OUT STD_LOGIC;
		csr_flag	: OUT STD_LOGIC;
		csr_addr	: OUT STD_LOGIC_VECTOR(IMM_I_LEN DOWNTO 0);
		csr_op		: OUT STD_LOGIC_VECTOR(FUNCT3_LEN DOWNTO 0);
		illegal_inst	: OUT STD_LOGIC;
	);

END inst_decoder;

ARCHITECTURE behavioural OF inst_decoder IS

-- See RISC-V Unprivileged ISA Documentation, section 2.3 for definitions
SIGNAL op_code	: STD_LOGIC_VECTOR(OPCODE_LEN	DOWNTO 0);
SIGNAL r_dest	: STD_LOGIC_VECTOR(RD_LEN	DOWNTO 0);
SIGNAL funct3	: STD_LOGIC_VECTOR(FUNCT3_LEN	DOWNTO 0);
SIGNAL r1	: STD_LOGIC_VECTOR(R1_LEN	DOWNTO 0);
SIGNAL r2	: STD_LOGIC_VECTOR(R2_LEN	DOWNTO 0);
SIGNAL funct7	: STD_LOGIC_VECTOR(FUNCT7_LEN	DOWNTO 0);
SIGNAL funct2   : STD_LOGIC_VECTOR(FUNCT2_LEN	DOWNTO 0);
SIGNAL r3	: STD_LOGIC_VECTOR(R3_LEN	DOWNTO 0);
SIGNAL imm_i	: STD_LOGIC_VECTOR(IMM_I_LEN	DOWNTO 0);
SIGNAL imm_u	: STD_LOGIC_VECTOR(IMM_U_LEN	DOWNTO 0);
SIGNAL imm_s	: STD_LOGIC_VECTOR(IMM_S_LEN	DOWNTO 0);
SIGNAL imm_j	: STD_LOGIC_VECTOR(IMM_J_LEN	DOWNTO 0);
SIGNAL imm_b    : STD_LOGIC_VECTOR(IMM_B_LEN	DOWNTO 0);
SIGNAL funct4   : STD_LOGIC_VECTOR(FUNCT4_LEN   DOWNTO 0);
-- See definition in Unprivileged ISA doc, section 2.7
SIGNAL fence_pred_sig: STD_LOGIC_VECTOR(FENCE_PRED_LEN DOWNTO 0);
SIGNAL fence_succ_sig: STD_LOGIC_VECTOR(FENCE_SUCC_LEN DOWNTO 0);

BEGIN

--See Unpriv ISA doc, section 2.3
op_code <= instruction(OPCODE_START	DOWNTO OPCODE_END_2);
r_dest	<= instruction(RD_START		DOWNTO RD_END);
funct3	<= instruction(FUNCT3_START	DOWNTO FUNCT3_END);
r1	<= instruction(R1_START		DOWNTO R1_END);
r2	<= instruction(R2_START		DOWNTO R2_END);
funct7	<= instruction(FUNCT7_START	DOWNTO FUNCT7_END);
funct2  <= instruction(FUNCT2_START	DOWNTO FUNCT2_END);
funct4  <= instruction(FUNCT4_START	DOWNTO FUNCT4_END);
fence_pred_sig <= instruction(FENCE_PRED_START DOWNTO FENCE_PRED_END);
fence_succ_sig <= instruction(FENCE_SUCC_START DOWNTO FENCE_SUCC_END);
r3	<= instruction(R3_START		DOWNTO R3_END);
imm_i	<= instruction(IMM_I_START	DOWNTO IMM_I_END);
imm_u	<= instruction(IMM_U_START	DOWNTO IMM_U_END);
imm_s   <= instruction(IMM_S_A_START	DOWNTO IMM_S_A_END) & instruction(IMM_S_B_START	DOWNTO IMM_S_B_END);
imm_j   <= instruction(31) & instruction(19 DOWNTO 12) & instruction(20) & instruction(30 DOWNTO 21);
imm_b   <= instruction(31) & instruction(7) & instruction(30 DOWNTO 25) & instruction(11 DOWNTO 8);

--Export internal signals
destination <= r_dest;
reg1 <= r1;
reg2 <= r2;

-- Main behavioural block
op_code_inst : PROCESS(clk) BEGIN
	IF (rising_edge(clk)) THEN
		--Set all flags to 0
		invalid_inst <= '0';
		xload_flag <= '0';
		fload_flag <= '0';
		opimm_flag <= '0';
		auipc_flag <= '0';
		lui_flag   <= '0';
		amo_flag   <= '0';
		xstore_flag<= '0';
		fstore_flag<= '0';
		op_flag	   <= '0';
		flop_flag <= '0';
		hflop_flag<= '0';
		-- Opcodes must have the lowest 2 bits be '11' 
		IF (NOT("11" = instruction(1 downto 0))) THEN
			invalid_inst <= '1';
			exit;
		END IF;

		-- Match to the correct op-code
		CASE op_code IS
			WHEN OPCODE_LOAD	=>
				--Load to x reg
				xload_flag	<= '1';
				xload_offset	<= imm_i;
				CASE funct3 IS
					WHEN F3_LOAD_LW =>
						xload_mask <= WORD_MASK;
					WHEN F3_LOAD_LB =>
						xload_mask <= BYTE_MASK;
					WHEN F3_LOAD_LBU =>
						xload_mask <= BYTE_UPPER_MASK;
					WHEN F3_LOAD_LH =>
						xload_mask <= HALF_MASK;
					WHEN F3_LOAD_LHU =>
						xload_mask <= HALF_UPPER_MASK;
					WHEN OTHERS =>
						xload_flag <= '0';
				END CASE;
			WHEN OPCODE_FLOAD	=>
				IF (F2_LOAD_FLW = funct3) THEN
					--Load to float reg
					fload_flag	<= '1';
					fload_offset	<= imm_i;
				END IF;
			WHEN OPCODE_MISCMEM	=>
				--**************************************************
				--Memory ops; mostly fences
				--**************************************************
				IF (R_DEST_FENCE = r_dest AND R1_FENCE = r1) THEN
					IF (F4_FENCE = funct4) THEN
						fence_flag <= '1';
						fence_pred <= fence_pred_sig;
						fence_succ <= fence_succ_sig;
					ELSIF (F4_TSO = funct4) THEN
						tso_flag <= '1';
					END IF;
				END IF;
			WHEN OPCODE_OPIMM	=>
				--Immediate
				op_flag <= '1';
				immediate_flag <= '1';
				CASE funct3 IS
					WHEN F3_OP_ADD  =>
						op_op <= OP_ADD;
					WHEN F3_OP_SLT  =>
						op_op <= OP_SLT;
					WHEN F3_OP_SLTU =>
						op_op <= OP_SLTU;
					WHEN F3_OP_XOR  =>
						op_op <= OP_XOR;
					WHEN F3_OP_OR   =>
						op_op <= OP_OR;
					WHEN F3_OP_AND  =>
						op_op <= OP_AND;
					WHEN F3_OP_SLL  =>
						immediate <= (4 DOWNTO 0 => r2; OTHERS => '0');
						op_op <= OP_SLL;
					WHEN F3_OP_SR  =>
						--Shift Right Logical and Shift Right Arithmetic are
						-- the same F3
						IF(F7_OP_SRL = funct7) THEN
							op_op <= OP_SRL;
						ELSIF(F7_OP_SRA = funct7) THEN
							op_op <= OP_SRA;
					WHEN OTHERS =>
						op_flag <= '0';
				END CASE;
			WHEN OPCODE_AUIPC	=>
				--Add Upper Immedate to Program Counter
				auipc_flag <= '1';
				auipc_imm  <= imm_i;
			WHEN OPCODE_STORE	=>
				--Store x reg
				xstore_flag <= '1';
				xstore_imm  <= funct7 & r_dest;
				CASE funct3 IS
					WHEN F3_STORE_SB =>
						xstore_mask <= BYTE_MASK;
					WHEN F3_STORE_SH =>
						xstore_mask <= HALF_MASK;
					WHEN F3_STORE_SW =>
						xstore_mask <= WORD_MASK;
					WHEN OTHERS =>
						xstore_flag <= '0';
				END CASE;
			WHEN OPCODE_FSTORE	=>
				--Store float reg
				fstore_flag <= '1';
				fstore_imm  <= funct7 & r_dest;
				CASE funct3 IS
					WHEN F3_STORE_SB =>
						fstore_mask <= BYTE_MASK;
					WHEN F3_STORE_SH =>
						fstore_mask <= HALF_MASK;
					WHEN F3_STORE_SW =>
						fstore_mask <= WORD_MASK;
					WHEN OTHERS =>
						fstore_flag <= '0';
				END CASE;
			WHEN OPCODE_AMO		=>
				--Atomic Memory Operations
				IF (F3_AMO = funct3) THEN
					amo_flag <= '1';
					amo_funct7 <= amo_funct7;
				END IF;
			WHEN OPCODE_OP		=>
				--Misc Operations
				op_flag <= '1';
				IF (F3_OP_ADD = funct3 AND F7_OP_ADD = funct7) THEN op_op <= OP_ADD;
				ELSIF (F3_OP_SUB = funct3 AND F7_OP_SUB = funct7) THEN op_op <= OP_SUB;
				ELSIF (F3_OP_SLL = funct3 AND F7_OP_SLL = funct7) THEN op_op <= OP_SLL;
				ELSIF (F3_OP_SLT = funct3 AND F7_OP_SLT = funct7) THEN op_op <= OP_SLT;
				ELSIF (F3_OP_SLTU = funct3 AND F7_OP_SLTU = funct7) THEN op_op <= OP_SLTU;
				ELSIF (F3_OP_XOR = funct3 AND F7_OP_XOR = funct7) THEN op_op <= OP_XOR;
				ELSIF (F3_OP_SR = funct3 AND F7_OP_SRL = funct7) THEN op_op <= OP_SRL;
				ELSIF (F3_OP_SR = funct3 AND F7_OP_SRA = funct7) THEN op_op <= OP_SRA;
				ELSIF (F3_OP_OR = funct3 AND F7_OP_OR = funct7) THEN op_op <= OP_OR;
				ELSIF (F3_OP_AND = funct3 AND F7_OP_AND = funct7) THEN op_op <= OP_AND;
				ELSIF (F3_OP_M_MUL = funct3 AND F7_OP_M_EXT = funct7) THEN op_op <= OP_MUL;
				ELSIF (F3_OP_M_MULH = funct3 AND F7_OP_M_EXT = funct7) THEN op_op <= OP_MULH;
				ELSIF (F3_OP_M_MULHSU = funct3 AND F7_OP_M_EXT = funct7) THEN op_op <= OP_MULHSU;
				ELSIF (F3_OP_M_MULHU = funct3 AND F7_OP_M_EXT = funct7) THEN op_op <= OP_MULHU;
				ELSIF (F3_OP_M_DIV = funct3 AND F7_OP_M_EXT = funct7) THEN op_op <= OP_DIV;
				ELSIF (F3_OP_M_DIVU = funct3 AND F7_OP_M_EXT = funct7) THEN op_op <= OP_DIVU;
				ELSIF (F3_OP_M_REM = funct3 AND F7_OP_M_EXT = funct7) THEN op_op <= OP_REM;
				ELSIF (F3_OP_M_REMU = funct3 AND F7_OP_M_EXT = funct7) THEN op_op <= OP_REMU;
				ELSE
					op_flag <= '0';
				END IF;
			WHEN OPCODE_LUI		=>
				--Load Upper Immediate
				lui_flag <= '1';
				lui_imm  <= imm_i;
			WHEN OPCODE_MADD	=>
				--Multiply and Add
				IF (FLOAT_INST = funct2) THEN
					flop_flag <= '1';
				ELSIF (HFLOAT_INST = funct2) THEN 
					hflop_flag <= '1';
				END IF;
				flop_r3   <= r3;
				flop_rm   <= funct3;
				flop_op   <= FLOAT_OP_MADD;
			WHEN OPCODE_MSUB	=>
				--Multiply and Subtract
				IF (FLOAT_INST = funct2) THEN
					flop_flag <= '1';
				ELSIF (HFLOAT_INST = funct2) THEN 
					hflop_flag <= '1';
				END IF;
				flop_r3   <= r3;
				flop_rm   <= funct3;
				flop_op   <= FLOAT_OP_MSUB;
			WHEN OPCODE_NMADD	=>
				--Negate, multiply, add
				IF (FLOAT_INST = funct2) THEN
					flop_flag <= '1';
				ELSIF (HFLOAT_INST = funct2) THEN 
					hflop_flag <= '1';
				END IF;
				flop_r3   <= r3;
				flop_rm   <= funct3;
				flop_op   <= FLOAT_OP_NMADD;
			WHEN OPCODE_NMSUB	=>
				--Negate, multiply, subtract
				IF (FLOAT_INST = funct2) THEN
					flop_flag <= '1';
				ELSIF (HFLOAT_INST = funct2) THEN 
					hflop_flag <= '1';
				END IF;
				flop_r3   <= r3;
				flop_rm   <= funct3;
				flop_op   <= FLOAT_OP_NMSUB;
			WHEN OPCODE_FLOAT	=>
				--Float ops
				IF (FLOAT_INST = funct2) THEN
					flop_flag <= '1';
				ELSIF (HFLOAT_INST = funct2) THEN 
					hflop_flag <= '1';
				END IF;
				flop_rm   <= funct3;
				CASE r3 IS
					WHEN R3_OP_FADD => flop_op <= FLOAT_OP_ADD;
					WHEN R3_OP_FSUB => flop_op <= FLOAT_OP_SUB;
					WHEN R3_OP_FMUL => flop_op <= FLOAT_OP_MUL;
					WHEN R3_OP_FDIV => flop_op <= FLOAT_OP_DIV;
					WHEN R3_OP_FSQRT =>
						IF (R2_OP_FSQRT = r2) THEN
							flop_op <= FLOAT_OP_SQRT;
						ELSE
							flop_flag <= '0';
							hflop_flag <= '0';
						END IF;
					WHEN R3_OP_FSIGN =>
						CASE funct3 IS
							WHEN F3_OP_FSGNJ => 
								flop_op <= FLOAT_OP_SGNJ;
							WHEN F3_OP_FSGNJN => 
								flop_op <= FLOAT_OP_SGNJN;
							WHEN F3_OP_FSGNJX => 
								flop_op <= FLOAT_OP_SGNJX;
							WHEN OTHERS =>
								hflop_flag <= "0";
								flop_flag <= "0";
						END CASE;
					WHEN R3_OP_FMINMAX =>
						CASE funct3 IS
							WHEN F3_OP_FMIN => 
								flop_op <= FLOAT_OP_MIN;
							WHEN F3_OP_FMAX => 
								flop_op <= FLOAT_OP_MAX;
							WHEN OTHERS =>
								hflop_flag <= "0";
								flop_flag <= "0";
						END CASE;
					WHEN R3_OP_FCONVW =>
						CASE r2 IS
							WHEN R2_OP_FCVTW =>
								flop_op <= FLOAT_OP_CVTW;
							WHEN R2_OP_FCVTWU =>
								flop_op <= FLOAT_OP_CVTWU;
							WHEN OTHERS =>
								hflop_flag <= "0";
								flop_flag <= "0";
						END CASE;
					WHEN R3_OP_COMP =>
						CASE funct3 IS
							WHEN F3_OP_FEQ => 
								flop_op <= FLOP_OP_EQ;
							WHEN F3_OP_FLT =>
								flop_op <= FLOP_OP_LT;
							WHEN F3_OP_FLE =>
								flop_op <= FLOP_OP_LE;
							WHEN OTHERS =>
								hflop_flag <= "0";
								flop_flag <= "0";
						END CASE;
					WHEN R3_OP_CLASS =>
						CASE funct3 IS
							WHEN F3_OP_MVX => 
								flop_op <= FLOP_OP_MVX;
							WHEN F3_OP_CLASS =>
								flop_op <= FLOP_OP_CLASS;
							WHEN OTHERS =>
								hflop_flag <= "0";
								flop_flag <= "0";
						END CASE;
				END CASE;
			WHEN OPCODE_VECTOR	=>
				--Vector ops
			WHEN OPCODE_BRANCH	=>
				--Branch ops
				branch_flag <= '1';
				branch_offset <= imm_b;
				CASE funct3 IS
					WHEN F3_BRANCH_BEQ =>
						branch_op <= F3_BRANCH_BEQ;
					WHEN F3_BRANCH_BNE =>
						branch_op <= F3_BRANCH_BNE;
					WHEN F3_BRANCH_BLT =>
						branch_op <= F3_BRANCH_BLT;
					WHEN F3_BRANCH_BGE =>
						branch_op <= F3_BRANCH_BGE;
					WHEN F3_BRANCH_BLTU =>
						branch_op <= F3_BRANCH_BLTU;
					WHEN F3_BRANCH_BGEU =>
						branch_op <= F3_BRANCH_BGEU;
					WHEN OTHERS =>
						branch_flag <= '0';
				END CASE;
			WHEN OPCODE_JALR	=>
				--Jump and Link Register
				jal_flag <= '1';
				jal_offset <= imm_j;
			WHEN OPCODE_JAL		=>
				--Jump and Link
				IF("000" = funct3) THEN
					jalr_flag <= '1';
					jalr_offset <= imm_i;
				END IF;

			WHEN OPCODE_SYSTEM	=>
				csr_addr <= imm_i;
				csr_op <= funct3;
				--System Instructions
				CASE funct3 IS
					WHEN F3_SYSTEM_ECALL =>
						CASE imm_u IS
							WHEN IMM_U_ECALL => ecall_flag <= '1';
							WHEN IMM_U_EBREAK => ebreak_flag <= '1';
							WHEN IMM_U_NTO => nto_flag <= '1';
							WHEN IMM_U_STO => sto_flag <= '1';
						END CASE;
				--CSR instructions
					WHEN F3_SYSTEM_CSRRW | F3_SYSTEM_CSRRS | 
					     F3_SYSTEM_CSRRC | F3_SYSTEM_CSRRWI | 
					     F3_SYSTEM_CSRRSI | F3_SYSTEM_CSRRCI =>
						     csr_flag <= '1';
				END CASE;
		END CASE;
	END IF;
END PROCESS;

END behavioural;
